-- Generated By protoc-gen-lua Do not Edit
local protobuf = require "protobuf"
module('ws_message_pb')


local WEBSHIELDDATA = protobuf.Descriptor();
local WEBSHIELDDATA_CAPTCHA = protobuf.EnumDescriptor();
local WEBSHIELDDATA_CAPTCHA_NA_ENUM = protobuf.EnumValueDescriptor();
local WEBSHIELDDATA_CAPTCHA_REQUESTED_ENUM = protobuf.EnumValueDescriptor();
local WEBSHIELDDATA_CAPTCHA_PASSED_ENUM = protobuf.EnumValueDescriptor();
local WEBSHIELDDATA_CAPTCHA_FAILED_ENUM = protobuf.EnumValueDescriptor();
local WEBSHIELDDATA_USER_ID_FIELD = protobuf.FieldDescriptor();
local WEBSHIELDDATA_REQUEST_DATA_FIELD = protobuf.FieldDescriptor();
local WEBSHIELDDATA_MODSEC_DATA_FIELD = protobuf.FieldDescriptor();
local WEBSHIELDDATA_CAPTCHA_FIELD = protobuf.FieldDescriptor();
local REQUESTDATA = protobuf.Descriptor();
local REQUESTDATA_HEADER_FIELD = protobuf.FieldDescriptor();
local REQUESTDATA_METHOD_FIELD = protobuf.FieldDescriptor();
local REQUESTDATA_URI_FIELD = protobuf.FieldDescriptor();
local HEADERITEM = protobuf.Descriptor();
local HEADERITEM_NAME_FIELD = protobuf.FieldDescriptor();
local HEADERITEM_VALUE_FIELD = protobuf.FieldDescriptor();
local MODSECURITYDATA = protobuf.Descriptor();
local MODSECURITYDATA_ACCESS_DENIED_FIELD = protobuf.FieldDescriptor();
local MODSECURITYDATA_OPT_FIELD = protobuf.FieldDescriptor();
local MODSECURITYDATA_TAGS_FIELD = protobuf.FieldDescriptor();
local MODSECURITYOPTION = protobuf.Descriptor();
local MODSECURITYOPTION_NAME_FIELD = protobuf.FieldDescriptor();
local MODSECURITYOPTION_VALUE_FIELD = protobuf.FieldDescriptor();
local PROACTIVEQUEUE = protobuf.Descriptor();
local PROACTIVEQUEUE_ACTION = protobuf.EnumDescriptor();
local PROACTIVEQUEUE_ACTION_LOG_ENUM = protobuf.EnumValueDescriptor();
local PROACTIVEQUEUE_ACTION_KILL_ENUM = protobuf.EnumValueDescriptor();
local PROACTIVEQUEUE_ACTION_BLOCK_ENUM = protobuf.EnumValueDescriptor();
local PROACTIVEQUEUE_VERSION_FIELD = protobuf.FieldDescriptor();
local PROACTIVEQUEUE_RECOGNIZER_ID_FIELD = protobuf.FieldDescriptor();
local PROACTIVEQUEUE_RECOGNIZER_SUB_ID_FIELD = protobuf.FieldDescriptor();
local PROACTIVEQUEUE_ACTION_FIELD = protobuf.FieldDescriptor();
local PROACTIVEQUEUE_QUEUE_FIELD = protobuf.FieldDescriptor();
local PROACTIVEQUEUE_ENV_FIELD = protobuf.FieldDescriptor();
local SENDINFO = protobuf.Descriptor();
local SENDINFO_METHOD = protobuf.EnumDescriptor();
local SENDINFO_METHOD_PROACTIVE_ENUM = protobuf.EnumValueDescriptor();
local SENDINFO_METHOD_WEBSHIELD_ENUM = protobuf.EnumValueDescriptor();
local SENDINFO_METHOD_PAM_ENUM = protobuf.EnumValueDescriptor();
local SENDINFO_METHOD_FIELD = protobuf.FieldDescriptor();
local SENDINFO_IP_FIELD = protobuf.FieldDescriptor();
local SENDINFO_TIMESTAMP_FIELD = protobuf.FieldDescriptor();
local SENDINFO_WEBSH_FIELD = protobuf.FieldDescriptor();
local SENDINFO_PROACTIVE_FIELD = protobuf.FieldDescriptor();

WEBSHIELDDATA_CAPTCHA_NA_ENUM.name = "NA"
WEBSHIELDDATA_CAPTCHA_NA_ENUM.index = 0
WEBSHIELDDATA_CAPTCHA_NA_ENUM.number = 0
WEBSHIELDDATA_CAPTCHA_REQUESTED_ENUM.name = "REQUESTED"
WEBSHIELDDATA_CAPTCHA_REQUESTED_ENUM.index = 1
WEBSHIELDDATA_CAPTCHA_REQUESTED_ENUM.number = 1
WEBSHIELDDATA_CAPTCHA_PASSED_ENUM.name = "PASSED"
WEBSHIELDDATA_CAPTCHA_PASSED_ENUM.index = 2
WEBSHIELDDATA_CAPTCHA_PASSED_ENUM.number = 2
WEBSHIELDDATA_CAPTCHA_FAILED_ENUM.name = "FAILED"
WEBSHIELDDATA_CAPTCHA_FAILED_ENUM.index = 3
WEBSHIELDDATA_CAPTCHA_FAILED_ENUM.number = 3
WEBSHIELDDATA_CAPTCHA.name = "Captcha"
WEBSHIELDDATA_CAPTCHA.full_name = ".i360parcel.WebShieldData.Captcha"
WEBSHIELDDATA_CAPTCHA.values = {WEBSHIELDDATA_CAPTCHA_NA_ENUM,WEBSHIELDDATA_CAPTCHA_REQUESTED_ENUM,WEBSHIELDDATA_CAPTCHA_PASSED_ENUM,WEBSHIELDDATA_CAPTCHA_FAILED_ENUM}
WEBSHIELDDATA_USER_ID_FIELD.name = "user_id"
WEBSHIELDDATA_USER_ID_FIELD.full_name = ".i360parcel.WebShieldData.user_id"
WEBSHIELDDATA_USER_ID_FIELD.number = 1
WEBSHIELDDATA_USER_ID_FIELD.index = 0
WEBSHIELDDATA_USER_ID_FIELD.label = 1
WEBSHIELDDATA_USER_ID_FIELD.has_default_value = true
WEBSHIELDDATA_USER_ID_FIELD.default_value = "0000000000000000000000000000000000000000"
WEBSHIELDDATA_USER_ID_FIELD.type = 9
WEBSHIELDDATA_USER_ID_FIELD.cpp_type = 9

WEBSHIELDDATA_REQUEST_DATA_FIELD.name = "request_data"
WEBSHIELDDATA_REQUEST_DATA_FIELD.full_name = ".i360parcel.WebShieldData.request_data"
WEBSHIELDDATA_REQUEST_DATA_FIELD.number = 2
WEBSHIELDDATA_REQUEST_DATA_FIELD.index = 1
WEBSHIELDDATA_REQUEST_DATA_FIELD.label = 1
WEBSHIELDDATA_REQUEST_DATA_FIELD.has_default_value = false
WEBSHIELDDATA_REQUEST_DATA_FIELD.default_value = nil
WEBSHIELDDATA_REQUEST_DATA_FIELD.message_type = _REQUESTDATA
WEBSHIELDDATA_REQUEST_DATA_FIELD.type = 11
WEBSHIELDDATA_REQUEST_DATA_FIELD.cpp_type = 10

WEBSHIELDDATA_MODSEC_DATA_FIELD.name = "modsec_data"
WEBSHIELDDATA_MODSEC_DATA_FIELD.full_name = ".i360parcel.WebShieldData.modsec_data"
WEBSHIELDDATA_MODSEC_DATA_FIELD.number = 3
WEBSHIELDDATA_MODSEC_DATA_FIELD.index = 2
WEBSHIELDDATA_MODSEC_DATA_FIELD.label = 1
WEBSHIELDDATA_MODSEC_DATA_FIELD.has_default_value = false
WEBSHIELDDATA_MODSEC_DATA_FIELD.default_value = nil
WEBSHIELDDATA_MODSEC_DATA_FIELD.message_type = _MODSECURITYDATA
WEBSHIELDDATA_MODSEC_DATA_FIELD.type = 11
WEBSHIELDDATA_MODSEC_DATA_FIELD.cpp_type = 10

WEBSHIELDDATA_CAPTCHA_FIELD.name = "captcha"
WEBSHIELDDATA_CAPTCHA_FIELD.full_name = ".i360parcel.WebShieldData.captcha"
WEBSHIELDDATA_CAPTCHA_FIELD.number = 4
WEBSHIELDDATA_CAPTCHA_FIELD.index = 3
WEBSHIELDDATA_CAPTCHA_FIELD.label = 1
WEBSHIELDDATA_CAPTCHA_FIELD.has_default_value = false
WEBSHIELDDATA_CAPTCHA_FIELD.default_value = nil
WEBSHIELDDATA_CAPTCHA_FIELD.enum_type = WEBSHIELDDATA_CAPTCHA
WEBSHIELDDATA_CAPTCHA_FIELD.type = 14
WEBSHIELDDATA_CAPTCHA_FIELD.cpp_type = 8

WEBSHIELDDATA.name = "WebShieldData"
WEBSHIELDDATA.full_name = ".i360parcel.WebShieldData"
WEBSHIELDDATA.nested_types = {}
WEBSHIELDDATA.enum_types = {WEBSHIELDDATA_CAPTCHA}
WEBSHIELDDATA.fields = {WEBSHIELDDATA_USER_ID_FIELD, WEBSHIELDDATA_REQUEST_DATA_FIELD, WEBSHIELDDATA_MODSEC_DATA_FIELD, WEBSHIELDDATA_CAPTCHA_FIELD}
WEBSHIELDDATA.is_extendable = false
WEBSHIELDDATA.extensions = {}
REQUESTDATA_HEADER_FIELD.name = "header"
REQUESTDATA_HEADER_FIELD.full_name = ".i360parcel.RequestData.header"
REQUESTDATA_HEADER_FIELD.number = 1
REQUESTDATA_HEADER_FIELD.index = 0
REQUESTDATA_HEADER_FIELD.label = 3
REQUESTDATA_HEADER_FIELD.has_default_value = false
REQUESTDATA_HEADER_FIELD.default_value = {}
REQUESTDATA_HEADER_FIELD.message_type = _HEADERITEM
REQUESTDATA_HEADER_FIELD.type = 11
REQUESTDATA_HEADER_FIELD.cpp_type = 10

REQUESTDATA_METHOD_FIELD.name = "method"
REQUESTDATA_METHOD_FIELD.full_name = ".i360parcel.RequestData.method"
REQUESTDATA_METHOD_FIELD.number = 2
REQUESTDATA_METHOD_FIELD.index = 1
REQUESTDATA_METHOD_FIELD.label = 1
REQUESTDATA_METHOD_FIELD.has_default_value = true
REQUESTDATA_METHOD_FIELD.default_value = "GET"
REQUESTDATA_METHOD_FIELD.type = 9
REQUESTDATA_METHOD_FIELD.cpp_type = 9

REQUESTDATA_URI_FIELD.name = "uri"
REQUESTDATA_URI_FIELD.full_name = ".i360parcel.RequestData.uri"
REQUESTDATA_URI_FIELD.number = 3
REQUESTDATA_URI_FIELD.index = 2
REQUESTDATA_URI_FIELD.label = 1
REQUESTDATA_URI_FIELD.has_default_value = false
REQUESTDATA_URI_FIELD.default_value = ""
REQUESTDATA_URI_FIELD.type = 9
REQUESTDATA_URI_FIELD.cpp_type = 9

REQUESTDATA.name = "RequestData"
REQUESTDATA.full_name = ".i360parcel.RequestData"
REQUESTDATA.nested_types = {}
REQUESTDATA.enum_types = {}
REQUESTDATA.fields = {REQUESTDATA_HEADER_FIELD, REQUESTDATA_METHOD_FIELD, REQUESTDATA_URI_FIELD}
REQUESTDATA.is_extendable = false
REQUESTDATA.extensions = {}
HEADERITEM_NAME_FIELD.name = "name"
HEADERITEM_NAME_FIELD.full_name = ".i360parcel.HeaderItem.name"
HEADERITEM_NAME_FIELD.number = 1
HEADERITEM_NAME_FIELD.index = 0
HEADERITEM_NAME_FIELD.label = 2
HEADERITEM_NAME_FIELD.has_default_value = false
HEADERITEM_NAME_FIELD.default_value = ""
HEADERITEM_NAME_FIELD.type = 9
HEADERITEM_NAME_FIELD.cpp_type = 9

HEADERITEM_VALUE_FIELD.name = "value"
HEADERITEM_VALUE_FIELD.full_name = ".i360parcel.HeaderItem.value"
HEADERITEM_VALUE_FIELD.number = 2
HEADERITEM_VALUE_FIELD.index = 1
HEADERITEM_VALUE_FIELD.label = 1
HEADERITEM_VALUE_FIELD.has_default_value = false
HEADERITEM_VALUE_FIELD.default_value = ""
HEADERITEM_VALUE_FIELD.type = 9
HEADERITEM_VALUE_FIELD.cpp_type = 9

HEADERITEM.name = "HeaderItem"
HEADERITEM.full_name = ".i360parcel.HeaderItem"
HEADERITEM.nested_types = {}
HEADERITEM.enum_types = {}
HEADERITEM.fields = {HEADERITEM_NAME_FIELD, HEADERITEM_VALUE_FIELD}
HEADERITEM.is_extendable = false
HEADERITEM.extensions = {}
MODSECURITYDATA_ACCESS_DENIED_FIELD.name = "access_denied"
MODSECURITYDATA_ACCESS_DENIED_FIELD.full_name = ".i360parcel.ModSecurityData.access_denied"
MODSECURITYDATA_ACCESS_DENIED_FIELD.number = 1
MODSECURITYDATA_ACCESS_DENIED_FIELD.index = 0
MODSECURITYDATA_ACCESS_DENIED_FIELD.label = 1
MODSECURITYDATA_ACCESS_DENIED_FIELD.has_default_value = false
MODSECURITYDATA_ACCESS_DENIED_FIELD.default_value = false
MODSECURITYDATA_ACCESS_DENIED_FIELD.type = 8
MODSECURITYDATA_ACCESS_DENIED_FIELD.cpp_type = 7

MODSECURITYDATA_OPT_FIELD.name = "opt"
MODSECURITYDATA_OPT_FIELD.full_name = ".i360parcel.ModSecurityData.opt"
MODSECURITYDATA_OPT_FIELD.number = 2
MODSECURITYDATA_OPT_FIELD.index = 1
MODSECURITYDATA_OPT_FIELD.label = 3
MODSECURITYDATA_OPT_FIELD.has_default_value = false
MODSECURITYDATA_OPT_FIELD.default_value = {}
MODSECURITYDATA_OPT_FIELD.message_type = _MODSECURITYOPTION
MODSECURITYDATA_OPT_FIELD.type = 11
MODSECURITYDATA_OPT_FIELD.cpp_type = 10

MODSECURITYDATA_TAGS_FIELD.name = "tags"
MODSECURITYDATA_TAGS_FIELD.full_name = ".i360parcel.ModSecurityData.tags"
MODSECURITYDATA_TAGS_FIELD.number = 3
MODSECURITYDATA_TAGS_FIELD.index = 2
MODSECURITYDATA_TAGS_FIELD.label = 3
MODSECURITYDATA_TAGS_FIELD.has_default_value = false
MODSECURITYDATA_TAGS_FIELD.default_value = {}
MODSECURITYDATA_TAGS_FIELD.type = 9
MODSECURITYDATA_TAGS_FIELD.cpp_type = 9

MODSECURITYDATA.name = "ModSecurityData"
MODSECURITYDATA.full_name = ".i360parcel.ModSecurityData"
MODSECURITYDATA.nested_types = {}
MODSECURITYDATA.enum_types = {}
MODSECURITYDATA.fields = {MODSECURITYDATA_ACCESS_DENIED_FIELD, MODSECURITYDATA_OPT_FIELD, MODSECURITYDATA_TAGS_FIELD}
MODSECURITYDATA.is_extendable = false
MODSECURITYDATA.extensions = {}
MODSECURITYOPTION_NAME_FIELD.name = "name"
MODSECURITYOPTION_NAME_FIELD.full_name = ".i360parcel.ModSecurityOption.name"
MODSECURITYOPTION_NAME_FIELD.number = 1
MODSECURITYOPTION_NAME_FIELD.index = 0
MODSECURITYOPTION_NAME_FIELD.label = 2
MODSECURITYOPTION_NAME_FIELD.has_default_value = false
MODSECURITYOPTION_NAME_FIELD.default_value = ""
MODSECURITYOPTION_NAME_FIELD.type = 9
MODSECURITYOPTION_NAME_FIELD.cpp_type = 9

MODSECURITYOPTION_VALUE_FIELD.name = "value"
MODSECURITYOPTION_VALUE_FIELD.full_name = ".i360parcel.ModSecurityOption.value"
MODSECURITYOPTION_VALUE_FIELD.number = 2
MODSECURITYOPTION_VALUE_FIELD.index = 1
MODSECURITYOPTION_VALUE_FIELD.label = 1
MODSECURITYOPTION_VALUE_FIELD.has_default_value = false
MODSECURITYOPTION_VALUE_FIELD.default_value = ""
MODSECURITYOPTION_VALUE_FIELD.type = 9
MODSECURITYOPTION_VALUE_FIELD.cpp_type = 9

MODSECURITYOPTION.name = "ModSecurityOption"
MODSECURITYOPTION.full_name = ".i360parcel.ModSecurityOption"
MODSECURITYOPTION.nested_types = {}
MODSECURITYOPTION.enum_types = {}
MODSECURITYOPTION.fields = {MODSECURITYOPTION_NAME_FIELD, MODSECURITYOPTION_VALUE_FIELD}
MODSECURITYOPTION.is_extendable = false
MODSECURITYOPTION.extensions = {}
PROACTIVEQUEUE_ACTION_LOG_ENUM.name = "LOG"
PROACTIVEQUEUE_ACTION_LOG_ENUM.index = 0
PROACTIVEQUEUE_ACTION_LOG_ENUM.number = 0
PROACTIVEQUEUE_ACTION_KILL_ENUM.name = "KILL"
PROACTIVEQUEUE_ACTION_KILL_ENUM.index = 1
PROACTIVEQUEUE_ACTION_KILL_ENUM.number = 1
PROACTIVEQUEUE_ACTION_BLOCK_ENUM.name = "BLOCK"
PROACTIVEQUEUE_ACTION_BLOCK_ENUM.index = 2
PROACTIVEQUEUE_ACTION_BLOCK_ENUM.number = 2
PROACTIVEQUEUE_ACTION.name = "Action"
PROACTIVEQUEUE_ACTION.full_name = ".i360parcel.ProactiveQueue.Action"
PROACTIVEQUEUE_ACTION.values = {PROACTIVEQUEUE_ACTION_LOG_ENUM,PROACTIVEQUEUE_ACTION_KILL_ENUM,PROACTIVEQUEUE_ACTION_BLOCK_ENUM}
PROACTIVEQUEUE_VERSION_FIELD.name = "version"
PROACTIVEQUEUE_VERSION_FIELD.full_name = ".i360parcel.ProactiveQueue.version"
PROACTIVEQUEUE_VERSION_FIELD.number = 1
PROACTIVEQUEUE_VERSION_FIELD.index = 0
PROACTIVEQUEUE_VERSION_FIELD.label = 2
PROACTIVEQUEUE_VERSION_FIELD.has_default_value = true
PROACTIVEQUEUE_VERSION_FIELD.default_value = 1
PROACTIVEQUEUE_VERSION_FIELD.type = 13
PROACTIVEQUEUE_VERSION_FIELD.cpp_type = 3

PROACTIVEQUEUE_RECOGNIZER_ID_FIELD.name = "recognizer_id"
PROACTIVEQUEUE_RECOGNIZER_ID_FIELD.full_name = ".i360parcel.ProactiveQueue.recognizer_id"
PROACTIVEQUEUE_RECOGNIZER_ID_FIELD.number = 2
PROACTIVEQUEUE_RECOGNIZER_ID_FIELD.index = 1
PROACTIVEQUEUE_RECOGNIZER_ID_FIELD.label = 2
PROACTIVEQUEUE_RECOGNIZER_ID_FIELD.has_default_value = false
PROACTIVEQUEUE_RECOGNIZER_ID_FIELD.default_value = 0
PROACTIVEQUEUE_RECOGNIZER_ID_FIELD.type = 13
PROACTIVEQUEUE_RECOGNIZER_ID_FIELD.cpp_type = 3

PROACTIVEQUEUE_RECOGNIZER_SUB_ID_FIELD.name = "recognizer_sub_id"
PROACTIVEQUEUE_RECOGNIZER_SUB_ID_FIELD.full_name = ".i360parcel.ProactiveQueue.recognizer_sub_id"
PROACTIVEQUEUE_RECOGNIZER_SUB_ID_FIELD.number = 30
PROACTIVEQUEUE_RECOGNIZER_SUB_ID_FIELD.index = 2
PROACTIVEQUEUE_RECOGNIZER_SUB_ID_FIELD.label = 1
PROACTIVEQUEUE_RECOGNIZER_SUB_ID_FIELD.has_default_value = false
PROACTIVEQUEUE_RECOGNIZER_SUB_ID_FIELD.default_value = 0
PROACTIVEQUEUE_RECOGNIZER_SUB_ID_FIELD.type = 13
PROACTIVEQUEUE_RECOGNIZER_SUB_ID_FIELD.cpp_type = 3

PROACTIVEQUEUE_ACTION_FIELD.name = "action"
PROACTIVEQUEUE_ACTION_FIELD.full_name = ".i360parcel.ProactiveQueue.action"
PROACTIVEQUEUE_ACTION_FIELD.number = 31
PROACTIVEQUEUE_ACTION_FIELD.index = 3
PROACTIVEQUEUE_ACTION_FIELD.label = 1
PROACTIVEQUEUE_ACTION_FIELD.has_default_value = false
PROACTIVEQUEUE_ACTION_FIELD.default_value = nil
PROACTIVEQUEUE_ACTION_FIELD.enum_type = PROACTIVEQUEUE_ACTION
PROACTIVEQUEUE_ACTION_FIELD.type = 14
PROACTIVEQUEUE_ACTION_FIELD.cpp_type = 8

PROACTIVEQUEUE_QUEUE_FIELD.name = "queue"
PROACTIVEQUEUE_QUEUE_FIELD.full_name = ".i360parcel.ProactiveQueue.queue"
PROACTIVEQUEUE_QUEUE_FIELD.number = 4
PROACTIVEQUEUE_QUEUE_FIELD.index = 4
PROACTIVEQUEUE_QUEUE_FIELD.label = 3
PROACTIVEQUEUE_QUEUE_FIELD.has_default_value = false
PROACTIVEQUEUE_QUEUE_FIELD.default_value = {}
PROACTIVEQUEUE_QUEUE_FIELD.type = 7
PROACTIVEQUEUE_QUEUE_FIELD.cpp_type = 3

PROACTIVEQUEUE_ENV_FIELD.name = "env"
PROACTIVEQUEUE_ENV_FIELD.full_name = ".i360parcel.ProactiveQueue.env"
PROACTIVEQUEUE_ENV_FIELD.number = 5
PROACTIVEQUEUE_ENV_FIELD.index = 5
PROACTIVEQUEUE_ENV_FIELD.label = 3
PROACTIVEQUEUE_ENV_FIELD.has_default_value = false
PROACTIVEQUEUE_ENV_FIELD.default_value = {}
PROACTIVEQUEUE_ENV_FIELD.type = 9
PROACTIVEQUEUE_ENV_FIELD.cpp_type = 9

PROACTIVEQUEUE.name = "ProactiveQueue"
PROACTIVEQUEUE.full_name = ".i360parcel.ProactiveQueue"
PROACTIVEQUEUE.nested_types = {}
PROACTIVEQUEUE.enum_types = {PROACTIVEQUEUE_ACTION}
PROACTIVEQUEUE.fields = {PROACTIVEQUEUE_VERSION_FIELD, PROACTIVEQUEUE_RECOGNIZER_ID_FIELD, PROACTIVEQUEUE_RECOGNIZER_SUB_ID_FIELD, PROACTIVEQUEUE_ACTION_FIELD, PROACTIVEQUEUE_QUEUE_FIELD, PROACTIVEQUEUE_ENV_FIELD}
PROACTIVEQUEUE.is_extendable = false
PROACTIVEQUEUE.extensions = {}
SENDINFO_METHOD_PROACTIVE_ENUM.name = "PROACTIVE"
SENDINFO_METHOD_PROACTIVE_ENUM.index = 0
SENDINFO_METHOD_PROACTIVE_ENUM.number = 0
SENDINFO_METHOD_WEBSHIELD_ENUM.name = "WEBSHIELD"
SENDINFO_METHOD_WEBSHIELD_ENUM.index = 1
SENDINFO_METHOD_WEBSHIELD_ENUM.number = 1
SENDINFO_METHOD_PAM_ENUM.name = "PAM"
SENDINFO_METHOD_PAM_ENUM.index = 2
SENDINFO_METHOD_PAM_ENUM.number = 2
SENDINFO_METHOD.name = "Method"
SENDINFO_METHOD.full_name = ".i360parcel.SendInfo.Method"
SENDINFO_METHOD.values = {SENDINFO_METHOD_PROACTIVE_ENUM,SENDINFO_METHOD_WEBSHIELD_ENUM,SENDINFO_METHOD_PAM_ENUM}
SENDINFO_METHOD_FIELD.name = "method"
SENDINFO_METHOD_FIELD.full_name = ".i360parcel.SendInfo.method"
SENDINFO_METHOD_FIELD.number = 1
SENDINFO_METHOD_FIELD.index = 0
SENDINFO_METHOD_FIELD.label = 2
SENDINFO_METHOD_FIELD.has_default_value = false
SENDINFO_METHOD_FIELD.default_value = nil
SENDINFO_METHOD_FIELD.enum_type = SENDINFO_METHOD
SENDINFO_METHOD_FIELD.type = 14
SENDINFO_METHOD_FIELD.cpp_type = 8

SENDINFO_IP_FIELD.name = "ip"
SENDINFO_IP_FIELD.full_name = ".i360parcel.SendInfo.ip"
SENDINFO_IP_FIELD.number = 2
SENDINFO_IP_FIELD.index = 1
SENDINFO_IP_FIELD.label = 2
SENDINFO_IP_FIELD.has_default_value = false
SENDINFO_IP_FIELD.default_value = ""
SENDINFO_IP_FIELD.type = 9
SENDINFO_IP_FIELD.cpp_type = 9

SENDINFO_TIMESTAMP_FIELD.name = "timestamp"
SENDINFO_TIMESTAMP_FIELD.full_name = ".i360parcel.SendInfo.timestamp"
SENDINFO_TIMESTAMP_FIELD.number = 4
SENDINFO_TIMESTAMP_FIELD.index = 2
SENDINFO_TIMESTAMP_FIELD.label = 2
SENDINFO_TIMESTAMP_FIELD.has_default_value = false
SENDINFO_TIMESTAMP_FIELD.default_value = 0
SENDINFO_TIMESTAMP_FIELD.type = 5
SENDINFO_TIMESTAMP_FIELD.cpp_type = 1

SENDINFO_WEBSH_FIELD.name = "websh"
SENDINFO_WEBSH_FIELD.full_name = ".i360parcel.SendInfo.websh"
SENDINFO_WEBSH_FIELD.number = 5
SENDINFO_WEBSH_FIELD.index = 3
SENDINFO_WEBSH_FIELD.label = 1
SENDINFO_WEBSH_FIELD.has_default_value = false
SENDINFO_WEBSH_FIELD.default_value = nil
SENDINFO_WEBSH_FIELD.message_type = WEBSHIELDDATA
SENDINFO_WEBSH_FIELD.type = 11
SENDINFO_WEBSH_FIELD.cpp_type = 10

SENDINFO_PROACTIVE_FIELD.name = "proactive"
SENDINFO_PROACTIVE_FIELD.full_name = ".i360parcel.SendInfo.proactive"
SENDINFO_PROACTIVE_FIELD.number = 6
SENDINFO_PROACTIVE_FIELD.index = 4
SENDINFO_PROACTIVE_FIELD.label = 1
SENDINFO_PROACTIVE_FIELD.has_default_value = false
SENDINFO_PROACTIVE_FIELD.default_value = nil
SENDINFO_PROACTIVE_FIELD.message_type = PROACTIVEQUEUE
SENDINFO_PROACTIVE_FIELD.type = 11
SENDINFO_PROACTIVE_FIELD.cpp_type = 10

SENDINFO.name = "SendInfo"
SENDINFO.full_name = ".i360parcel.SendInfo"
SENDINFO.nested_types = {}
SENDINFO.enum_types = {SENDINFO_METHOD}
SENDINFO.fields = {SENDINFO_METHOD_FIELD, SENDINFO_IP_FIELD, SENDINFO_TIMESTAMP_FIELD, SENDINFO_WEBSH_FIELD, SENDINFO_PROACTIVE_FIELD}
SENDINFO.is_extendable = false
SENDINFO.extensions = {}

HeaderItem = protobuf.Message(HEADERITEM)
ModSecurityData = protobuf.Message(MODSECURITYDATA)
ModSecurityOption = protobuf.Message(MODSECURITYOPTION)
ProactiveQueue = protobuf.Message(PROACTIVEQUEUE)
RequestData = protobuf.Message(REQUESTDATA)
SendInfo = protobuf.Message(SENDINFO)
WebShieldData = protobuf.Message(WEBSHIELDDATA)

