-- Generated By protoc-gen-lua Do not Edit
local protobuf = require "protobuf"
module('ipset_message_pb')


local HEADER = protobuf.Descriptor();
local HEADER_NAME_FIELD = protobuf.FieldDescriptor();
local HEADER_VALUE_FIELD = protobuf.FieldDescriptor();
local REQUEST = protobuf.Descriptor();
local REQUEST_IP_FIELD = protobuf.FieldDescriptor();
local REQUEST_UA_FIELD = protobuf.FieldDescriptor();
local REQUEST_URL_FIELD = protobuf.FieldDescriptor();
local REQUEST_UID_FIELD = protobuf.FieldDescriptor();
local REQUEST_HEADER_FIELD = protobuf.FieldDescriptor();
local REQUEST_CAPTCHA_FIELD = protobuf.FieldDescriptor();
local REQUEST_HOST_FIELD = protobuf.FieldDescriptor();
local REQUEST_LOCALE_FIELD = protobuf.FieldDescriptor();
local REQUEST_PORT_FIELD = protobuf.FieldDescriptor();
local REQUEST_PROXYNAME_FIELD = protobuf.FieldDescriptor();
local REQUEST_REALIP_FIELD = protobuf.FieldDescriptor();
local REQUEST_SPLASHCAPTCHA_FIELD = protobuf.FieldDescriptor();
local REQUEST_WLDOMAIN_FIELD = protobuf.FieldDescriptor();
local SPLASHREDIRECT = protobuf.Descriptor();
local SPLASHREDIRECT_LOCATION_FIELD = protobuf.FieldDescriptor();
local SPLASHREDIRECT_COOKIE_FIELD = protobuf.FieldDescriptor();
local RESPONSE = protobuf.Descriptor();
local RESPONSE_STATUS = protobuf.EnumDescriptor();
local RESPONSE_STATUS_OK_ENUM = protobuf.EnumValueDescriptor();
local RESPONSE_STATUS_FAIL_ENUM = protobuf.EnumValueDescriptor();
local RESPONSE_ACTION = protobuf.EnumDescriptor();
local RESPONSE_ACTION_NOTFOUND_ENUM = protobuf.EnumValueDescriptor();
local RESPONSE_ACTION_PASS_ENUM = protobuf.EnumValueDescriptor();
local RESPONSE_ACTION_WHITE_ENUM = protobuf.EnumValueDescriptor();
local RESPONSE_ACTION_REMOTE_PROXY_ENUM = protobuf.EnumValueDescriptor();
local RESPONSE_ACTION_WHITE_STATIC_ENUM = protobuf.EnumValueDescriptor();
local RESPONSE_ACTION_DENY_ENUM = protobuf.EnumValueDescriptor();
local RESPONSE_ACTION_COUNTRY_BLACK_DENY_ENUM = protobuf.EnumValueDescriptor();
local RESPONSE_ACTION_DENY_LOCAL_ENUM = protobuf.EnumValueDescriptor();
local RESPONSE_ACTION_GRAYLIST_ENUM = protobuf.EnumValueDescriptor();
local RESPONSE_ACTION_GRAYLIST_PASS_ENUM = protobuf.EnumValueDescriptor();
local RESPONSE_ACTION_UNSUPPORTED_TYPE_ENUM = protobuf.EnumValueDescriptor();
local RESPONSE_ACTION_SPLASH_ENUM = protobuf.EnumValueDescriptor();
local RESPONSE_ACTION_SPLASH_PASS_ENUM = protobuf.EnumValueDescriptor();
local RESPONSE_ACTION_BADREQ_ENUM = protobuf.EnumValueDescriptor();
local RESPONSE_IP_FIELD = protobuf.FieldDescriptor();
local RESPONSE_STATUS_FIELD = protobuf.FieldDescriptor();
local RESPONSE_ACTION_FIELD = protobuf.FieldDescriptor();
local RESPONSE_SPLASHSCREEN_FIELD = protobuf.FieldDescriptor();
local RESPONSE_SPLASHREDIRECT_FIELD = protobuf.FieldDescriptor();
local RESPONSE_REALIP_FIELD = protobuf.FieldDescriptor();
local RESPONSE_HEADER_FIELD = protobuf.FieldDescriptor();
local RESPONSE_CAPTCHAPAGE_FIELD = protobuf.FieldDescriptor();
local RESPONSE_INTERNALHEADERS_FIELD = protobuf.FieldDescriptor();

HEADER_NAME_FIELD.name = "name"
HEADER_NAME_FIELD.full_name = ".msg.Header.name"
HEADER_NAME_FIELD.number = 1
HEADER_NAME_FIELD.index = 0
HEADER_NAME_FIELD.label = 2
HEADER_NAME_FIELD.has_default_value = false
HEADER_NAME_FIELD.default_value = ""
HEADER_NAME_FIELD.type = 9
HEADER_NAME_FIELD.cpp_type = 9

HEADER_VALUE_FIELD.name = "value"
HEADER_VALUE_FIELD.full_name = ".msg.Header.value"
HEADER_VALUE_FIELD.number = 2
HEADER_VALUE_FIELD.index = 1
HEADER_VALUE_FIELD.label = 1
HEADER_VALUE_FIELD.has_default_value = false
HEADER_VALUE_FIELD.default_value = ""
HEADER_VALUE_FIELD.type = 9
HEADER_VALUE_FIELD.cpp_type = 9

HEADER.name = "Header"
HEADER.full_name = ".msg.Header"
HEADER.nested_types = {}
HEADER.enum_types = {}
HEADER.fields = {HEADER_NAME_FIELD, HEADER_VALUE_FIELD}
HEADER.is_extendable = false
HEADER.extensions = {}
REQUEST_IP_FIELD.name = "ip"
REQUEST_IP_FIELD.full_name = ".msg.Request.ip"
REQUEST_IP_FIELD.number = 1
REQUEST_IP_FIELD.index = 0
REQUEST_IP_FIELD.label = 2
REQUEST_IP_FIELD.has_default_value = false
REQUEST_IP_FIELD.default_value = ""
REQUEST_IP_FIELD.type = 9
REQUEST_IP_FIELD.cpp_type = 9

REQUEST_UA_FIELD.name = "ua"
REQUEST_UA_FIELD.full_name = ".msg.Request.ua"
REQUEST_UA_FIELD.number = 2
REQUEST_UA_FIELD.index = 1
REQUEST_UA_FIELD.label = 2
REQUEST_UA_FIELD.has_default_value = false
REQUEST_UA_FIELD.default_value = ""
REQUEST_UA_FIELD.type = 9
REQUEST_UA_FIELD.cpp_type = 9

REQUEST_URL_FIELD.name = "url"
REQUEST_URL_FIELD.full_name = ".msg.Request.url"
REQUEST_URL_FIELD.number = 3
REQUEST_URL_FIELD.index = 2
REQUEST_URL_FIELD.label = 2
REQUEST_URL_FIELD.has_default_value = false
REQUEST_URL_FIELD.default_value = ""
REQUEST_URL_FIELD.type = 9
REQUEST_URL_FIELD.cpp_type = 9

REQUEST_UID_FIELD.name = "uid"
REQUEST_UID_FIELD.full_name = ".msg.Request.uid"
REQUEST_UID_FIELD.number = 4
REQUEST_UID_FIELD.index = 3
REQUEST_UID_FIELD.label = 2
REQUEST_UID_FIELD.has_default_value = false
REQUEST_UID_FIELD.default_value = ""
REQUEST_UID_FIELD.type = 9
REQUEST_UID_FIELD.cpp_type = 9

REQUEST_HEADER_FIELD.name = "header"
REQUEST_HEADER_FIELD.full_name = ".msg.Request.header"
REQUEST_HEADER_FIELD.number = 5
REQUEST_HEADER_FIELD.index = 4
REQUEST_HEADER_FIELD.label = 3
REQUEST_HEADER_FIELD.has_default_value = false
REQUEST_HEADER_FIELD.default_value = {}
REQUEST_HEADER_FIELD.message_type = HEADER
REQUEST_HEADER_FIELD.type = 11
REQUEST_HEADER_FIELD.cpp_type = 10

REQUEST_CAPTCHA_FIELD.name = "captcha"
REQUEST_CAPTCHA_FIELD.full_name = ".msg.Request.captcha"
REQUEST_CAPTCHA_FIELD.number = 6
REQUEST_CAPTCHA_FIELD.index = 5
REQUEST_CAPTCHA_FIELD.label = 2
REQUEST_CAPTCHA_FIELD.has_default_value = false
REQUEST_CAPTCHA_FIELD.default_value = ""
REQUEST_CAPTCHA_FIELD.type = 9
REQUEST_CAPTCHA_FIELD.cpp_type = 9

REQUEST_HOST_FIELD.name = "host"
REQUEST_HOST_FIELD.full_name = ".msg.Request.host"
REQUEST_HOST_FIELD.number = 7
REQUEST_HOST_FIELD.index = 6
REQUEST_HOST_FIELD.label = 2
REQUEST_HOST_FIELD.has_default_value = false
REQUEST_HOST_FIELD.default_value = ""
REQUEST_HOST_FIELD.type = 9
REQUEST_HOST_FIELD.cpp_type = 9

REQUEST_LOCALE_FIELD.name = "locale"
REQUEST_LOCALE_FIELD.full_name = ".msg.Request.locale"
REQUEST_LOCALE_FIELD.number = 8
REQUEST_LOCALE_FIELD.index = 7
REQUEST_LOCALE_FIELD.label = 1
REQUEST_LOCALE_FIELD.has_default_value = false
REQUEST_LOCALE_FIELD.default_value = ""
REQUEST_LOCALE_FIELD.type = 9
REQUEST_LOCALE_FIELD.cpp_type = 9

REQUEST_PORT_FIELD.name = "port"
REQUEST_PORT_FIELD.full_name = ".msg.Request.port"
REQUEST_PORT_FIELD.number = 9
REQUEST_PORT_FIELD.index = 8
REQUEST_PORT_FIELD.label = 1
REQUEST_PORT_FIELD.has_default_value = false
REQUEST_PORT_FIELD.default_value = ""
REQUEST_PORT_FIELD.type = 9
REQUEST_PORT_FIELD.cpp_type = 9

REQUEST_PROXYNAME_FIELD.name = "proxyname"
REQUEST_PROXYNAME_FIELD.full_name = ".msg.Request.proxyname"
REQUEST_PROXYNAME_FIELD.number = 10
REQUEST_PROXYNAME_FIELD.index = 9
REQUEST_PROXYNAME_FIELD.label = 1
REQUEST_PROXYNAME_FIELD.has_default_value = false
REQUEST_PROXYNAME_FIELD.default_value = ""
REQUEST_PROXYNAME_FIELD.type = 9
REQUEST_PROXYNAME_FIELD.cpp_type = 9

REQUEST_REALIP_FIELD.name = "realip"
REQUEST_REALIP_FIELD.full_name = ".msg.Request.realip"
REQUEST_REALIP_FIELD.number = 11
REQUEST_REALIP_FIELD.index = 10
REQUEST_REALIP_FIELD.label = 1
REQUEST_REALIP_FIELD.has_default_value = false
REQUEST_REALIP_FIELD.default_value = ""
REQUEST_REALIP_FIELD.type = 9
REQUEST_REALIP_FIELD.cpp_type = 9

REQUEST_SPLASHCAPTCHA_FIELD.name = "splashcaptcha"
REQUEST_SPLASHCAPTCHA_FIELD.full_name = ".msg.Request.splashcaptcha"
REQUEST_SPLASHCAPTCHA_FIELD.number = 12
REQUEST_SPLASHCAPTCHA_FIELD.index = 11
REQUEST_SPLASHCAPTCHA_FIELD.label = 1
REQUEST_SPLASHCAPTCHA_FIELD.has_default_value = false
REQUEST_SPLASHCAPTCHA_FIELD.default_value = false
REQUEST_SPLASHCAPTCHA_FIELD.type = 8
REQUEST_SPLASHCAPTCHA_FIELD.cpp_type = 7

REQUEST_WLDOMAIN_FIELD.name = "wldomain"
REQUEST_WLDOMAIN_FIELD.full_name = ".msg.Request.wldomain"
REQUEST_WLDOMAIN_FIELD.number = 13
REQUEST_WLDOMAIN_FIELD.index = 12
REQUEST_WLDOMAIN_FIELD.label = 1
REQUEST_WLDOMAIN_FIELD.has_default_value = false
REQUEST_WLDOMAIN_FIELD.default_value = false
REQUEST_WLDOMAIN_FIELD.type = 8
REQUEST_WLDOMAIN_FIELD.cpp_type = 7

REQUEST.name = "Request"
REQUEST.full_name = ".msg.Request"
REQUEST.nested_types = {}
REQUEST.enum_types = {}
REQUEST.fields = {REQUEST_IP_FIELD, REQUEST_UA_FIELD, REQUEST_URL_FIELD, REQUEST_UID_FIELD, REQUEST_HEADER_FIELD, REQUEST_CAPTCHA_FIELD, REQUEST_HOST_FIELD, REQUEST_LOCALE_FIELD, REQUEST_PORT_FIELD, REQUEST_PROXYNAME_FIELD, REQUEST_REALIP_FIELD, REQUEST_SPLASHCAPTCHA_FIELD, REQUEST_WLDOMAIN_FIELD}
REQUEST.is_extendable = false
REQUEST.extensions = {}
SPLASHREDIRECT_LOCATION_FIELD.name = "location"
SPLASHREDIRECT_LOCATION_FIELD.full_name = ".msg.SplashRedirect.location"
SPLASHREDIRECT_LOCATION_FIELD.number = 1
SPLASHREDIRECT_LOCATION_FIELD.index = 0
SPLASHREDIRECT_LOCATION_FIELD.label = 2
SPLASHREDIRECT_LOCATION_FIELD.has_default_value = false
SPLASHREDIRECT_LOCATION_FIELD.default_value = ""
SPLASHREDIRECT_LOCATION_FIELD.type = 9
SPLASHREDIRECT_LOCATION_FIELD.cpp_type = 9

SPLASHREDIRECT_COOKIE_FIELD.name = "cookie"
SPLASHREDIRECT_COOKIE_FIELD.full_name = ".msg.SplashRedirect.cookie"
SPLASHREDIRECT_COOKIE_FIELD.number = 2
SPLASHREDIRECT_COOKIE_FIELD.index = 1
SPLASHREDIRECT_COOKIE_FIELD.label = 2
SPLASHREDIRECT_COOKIE_FIELD.has_default_value = false
SPLASHREDIRECT_COOKIE_FIELD.default_value = ""
SPLASHREDIRECT_COOKIE_FIELD.type = 9
SPLASHREDIRECT_COOKIE_FIELD.cpp_type = 9

SPLASHREDIRECT.name = "SplashRedirect"
SPLASHREDIRECT.full_name = ".msg.SplashRedirect"
SPLASHREDIRECT.nested_types = {}
SPLASHREDIRECT.enum_types = {}
SPLASHREDIRECT.fields = {SPLASHREDIRECT_LOCATION_FIELD, SPLASHREDIRECT_COOKIE_FIELD}
SPLASHREDIRECT.is_extendable = false
SPLASHREDIRECT.extensions = {}
RESPONSE_STATUS_OK_ENUM.name = "OK"
RESPONSE_STATUS_OK_ENUM.index = 0
RESPONSE_STATUS_OK_ENUM.number = 0
RESPONSE_STATUS_FAIL_ENUM.name = "FAIL"
RESPONSE_STATUS_FAIL_ENUM.index = 1
RESPONSE_STATUS_FAIL_ENUM.number = 1
RESPONSE_STATUS.name = "Status"
RESPONSE_STATUS.full_name = ".msg.Response.Status"
RESPONSE_STATUS.values = {RESPONSE_STATUS_OK_ENUM,RESPONSE_STATUS_FAIL_ENUM}
RESPONSE_ACTION_NOTFOUND_ENUM.name = "NOTFOUND"
RESPONSE_ACTION_NOTFOUND_ENUM.index = 0
RESPONSE_ACTION_NOTFOUND_ENUM.number = 0
RESPONSE_ACTION_PASS_ENUM.name = "PASS"
RESPONSE_ACTION_PASS_ENUM.index = 1
RESPONSE_ACTION_PASS_ENUM.number = 1
RESPONSE_ACTION_WHITE_ENUM.name = "WHITE"
RESPONSE_ACTION_WHITE_ENUM.index = 2
RESPONSE_ACTION_WHITE_ENUM.number = 10
RESPONSE_ACTION_REMOTE_PROXY_ENUM.name = "REMOTE_PROXY"
RESPONSE_ACTION_REMOTE_PROXY_ENUM.index = 3
RESPONSE_ACTION_REMOTE_PROXY_ENUM.number = 11
RESPONSE_ACTION_WHITE_STATIC_ENUM.name = "WHITE_STATIC"
RESPONSE_ACTION_WHITE_STATIC_ENUM.index = 4
RESPONSE_ACTION_WHITE_STATIC_ENUM.number = 12
RESPONSE_ACTION_DENY_ENUM.name = "DENY"
RESPONSE_ACTION_DENY_ENUM.index = 5
RESPONSE_ACTION_DENY_ENUM.number = 20
RESPONSE_ACTION_COUNTRY_BLACK_DENY_ENUM.name = "COUNTRY_BLACK_DENY"
RESPONSE_ACTION_COUNTRY_BLACK_DENY_ENUM.index = 6
RESPONSE_ACTION_COUNTRY_BLACK_DENY_ENUM.number = 21
RESPONSE_ACTION_DENY_LOCAL_ENUM.name = "DENY_LOCAL"
RESPONSE_ACTION_DENY_LOCAL_ENUM.index = 7
RESPONSE_ACTION_DENY_LOCAL_ENUM.number = 22
RESPONSE_ACTION_GRAYLIST_ENUM.name = "GRAYLIST"
RESPONSE_ACTION_GRAYLIST_ENUM.index = 8
RESPONSE_ACTION_GRAYLIST_ENUM.number = 30
RESPONSE_ACTION_GRAYLIST_PASS_ENUM.name = "GRAYLIST_PASS"
RESPONSE_ACTION_GRAYLIST_PASS_ENUM.index = 9
RESPONSE_ACTION_GRAYLIST_PASS_ENUM.number = 31
RESPONSE_ACTION_UNSUPPORTED_TYPE_ENUM.name = "UNSUPPORTED_TYPE"
RESPONSE_ACTION_UNSUPPORTED_TYPE_ENUM.index = 10
RESPONSE_ACTION_UNSUPPORTED_TYPE_ENUM.number = 32
RESPONSE_ACTION_SPLASH_ENUM.name = "SPLASH"
RESPONSE_ACTION_SPLASH_ENUM.index = 11
RESPONSE_ACTION_SPLASH_ENUM.number = 40
RESPONSE_ACTION_SPLASH_PASS_ENUM.name = "SPLASH_PASS"
RESPONSE_ACTION_SPLASH_PASS_ENUM.index = 12
RESPONSE_ACTION_SPLASH_PASS_ENUM.number = 41
RESPONSE_ACTION_BADREQ_ENUM.name = "BADREQ"
RESPONSE_ACTION_BADREQ_ENUM.index = 13
RESPONSE_ACTION_BADREQ_ENUM.number = 50
RESPONSE_ACTION.name = "Action"
RESPONSE_ACTION.full_name = ".msg.Response.Action"
RESPONSE_ACTION.values = {RESPONSE_ACTION_NOTFOUND_ENUM,RESPONSE_ACTION_PASS_ENUM,RESPONSE_ACTION_WHITE_ENUM,RESPONSE_ACTION_REMOTE_PROXY_ENUM,RESPONSE_ACTION_WHITE_STATIC_ENUM,RESPONSE_ACTION_DENY_ENUM,RESPONSE_ACTION_COUNTRY_BLACK_DENY_ENUM,RESPONSE_ACTION_DENY_LOCAL_ENUM,RESPONSE_ACTION_GRAYLIST_ENUM,RESPONSE_ACTION_GRAYLIST_PASS_ENUM,RESPONSE_ACTION_UNSUPPORTED_TYPE_ENUM,RESPONSE_ACTION_SPLASH_ENUM,RESPONSE_ACTION_SPLASH_PASS_ENUM,RESPONSE_ACTION_BADREQ_ENUM}
RESPONSE_IP_FIELD.name = "ip"
RESPONSE_IP_FIELD.full_name = ".msg.Response.ip"
RESPONSE_IP_FIELD.number = 1
RESPONSE_IP_FIELD.index = 0
RESPONSE_IP_FIELD.label = 2
RESPONSE_IP_FIELD.has_default_value = false
RESPONSE_IP_FIELD.default_value = ""
RESPONSE_IP_FIELD.type = 9
RESPONSE_IP_FIELD.cpp_type = 9

RESPONSE_STATUS_FIELD.name = "status"
RESPONSE_STATUS_FIELD.full_name = ".msg.Response.status"
RESPONSE_STATUS_FIELD.number = 2
RESPONSE_STATUS_FIELD.index = 1
RESPONSE_STATUS_FIELD.label = 2
RESPONSE_STATUS_FIELD.has_default_value = false
RESPONSE_STATUS_FIELD.default_value = nil
RESPONSE_STATUS_FIELD.enum_type = RESPONSE_STATUS
RESPONSE_STATUS_FIELD.type = 14
RESPONSE_STATUS_FIELD.cpp_type = 8

RESPONSE_ACTION_FIELD.name = "action"
RESPONSE_ACTION_FIELD.full_name = ".msg.Response.action"
RESPONSE_ACTION_FIELD.number = 3
RESPONSE_ACTION_FIELD.index = 2
RESPONSE_ACTION_FIELD.label = 1
RESPONSE_ACTION_FIELD.has_default_value = false
RESPONSE_ACTION_FIELD.default_value = nil
RESPONSE_ACTION_FIELD.enum_type = RESPONSE_ACTION
RESPONSE_ACTION_FIELD.type = 14
RESPONSE_ACTION_FIELD.cpp_type = 8

RESPONSE_SPLASHSCREEN_FIELD.name = "splashscreen"
RESPONSE_SPLASHSCREEN_FIELD.full_name = ".msg.Response.splashscreen"
RESPONSE_SPLASHSCREEN_FIELD.number = 4
RESPONSE_SPLASHSCREEN_FIELD.index = 3
RESPONSE_SPLASHSCREEN_FIELD.label = 1
RESPONSE_SPLASHSCREEN_FIELD.has_default_value = false
RESPONSE_SPLASHSCREEN_FIELD.default_value = ""
RESPONSE_SPLASHSCREEN_FIELD.type = 9
RESPONSE_SPLASHSCREEN_FIELD.cpp_type = 9

RESPONSE_SPLASHREDIRECT_FIELD.name = "splashredirect"
RESPONSE_SPLASHREDIRECT_FIELD.full_name = ".msg.Response.splashredirect"
RESPONSE_SPLASHREDIRECT_FIELD.number = 5
RESPONSE_SPLASHREDIRECT_FIELD.index = 4
RESPONSE_SPLASHREDIRECT_FIELD.label = 1
RESPONSE_SPLASHREDIRECT_FIELD.has_default_value = false
RESPONSE_SPLASHREDIRECT_FIELD.default_value = nil
RESPONSE_SPLASHREDIRECT_FIELD.message_type = SPLASHREDIRECT
RESPONSE_SPLASHREDIRECT_FIELD.type = 11
RESPONSE_SPLASHREDIRECT_FIELD.cpp_type = 10

RESPONSE_REALIP_FIELD.name = "realIp"
RESPONSE_REALIP_FIELD.full_name = ".msg.Response.realIp"
RESPONSE_REALIP_FIELD.number = 6
RESPONSE_REALIP_FIELD.index = 5
RESPONSE_REALIP_FIELD.label = 1
RESPONSE_REALIP_FIELD.has_default_value = false
RESPONSE_REALIP_FIELD.default_value = ""
RESPONSE_REALIP_FIELD.type = 9
RESPONSE_REALIP_FIELD.cpp_type = 9

RESPONSE_HEADER_FIELD.name = "header"
RESPONSE_HEADER_FIELD.full_name = ".msg.Response.header"
RESPONSE_HEADER_FIELD.number = 7
RESPONSE_HEADER_FIELD.index = 6
RESPONSE_HEADER_FIELD.label = 3
RESPONSE_HEADER_FIELD.has_default_value = false
RESPONSE_HEADER_FIELD.default_value = {}
RESPONSE_HEADER_FIELD.message_type = HEADER
RESPONSE_HEADER_FIELD.type = 11
RESPONSE_HEADER_FIELD.cpp_type = 10

RESPONSE_CAPTCHAPAGE_FIELD.name = "captchapage"
RESPONSE_CAPTCHAPAGE_FIELD.full_name = ".msg.Response.captchapage"
RESPONSE_CAPTCHAPAGE_FIELD.number = 8
RESPONSE_CAPTCHAPAGE_FIELD.index = 7
RESPONSE_CAPTCHAPAGE_FIELD.label = 1
RESPONSE_CAPTCHAPAGE_FIELD.has_default_value = false
RESPONSE_CAPTCHAPAGE_FIELD.default_value = ""
RESPONSE_CAPTCHAPAGE_FIELD.type = 9
RESPONSE_CAPTCHAPAGE_FIELD.cpp_type = 9

RESPONSE_INTERNALHEADERS_FIELD.name = "internalHeaders"
RESPONSE_INTERNALHEADERS_FIELD.full_name = ".msg.Response.internalHeaders"
RESPONSE_INTERNALHEADERS_FIELD.number = 9
RESPONSE_INTERNALHEADERS_FIELD.index = 8
RESPONSE_INTERNALHEADERS_FIELD.label = 3
RESPONSE_INTERNALHEADERS_FIELD.has_default_value = false
RESPONSE_INTERNALHEADERS_FIELD.default_value = {}
RESPONSE_INTERNALHEADERS_FIELD.message_type = HEADER
RESPONSE_INTERNALHEADERS_FIELD.type = 11
RESPONSE_INTERNALHEADERS_FIELD.cpp_type = 10

RESPONSE.name = "Response"
RESPONSE.full_name = ".msg.Response"
RESPONSE.nested_types = {}
RESPONSE.enum_types = {RESPONSE_STATUS, RESPONSE_ACTION}
RESPONSE.fields = {RESPONSE_IP_FIELD, RESPONSE_STATUS_FIELD, RESPONSE_ACTION_FIELD, RESPONSE_SPLASHSCREEN_FIELD, RESPONSE_SPLASHREDIRECT_FIELD, RESPONSE_REALIP_FIELD, RESPONSE_HEADER_FIELD, RESPONSE_CAPTCHAPAGE_FIELD, RESPONSE_INTERNALHEADERS_FIELD}
RESPONSE.is_extendable = false
RESPONSE.extensions = {}

Header = protobuf.Message(HEADER)
Request = protobuf.Message(REQUEST)
Response = protobuf.Message(RESPONSE)
SplashRedirect = protobuf.Message(SPLASHREDIRECT)

